# (C) 2015 Elke Schaper

"""
    :synopsis: The RepeatList Class.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import logging
import pickle
import sys

LOG = logging.getLogger(__name__)

from tral.repeat_list import repeat_list_io


class RepeatList:

    """ A `RepeatList` contains list of repeats that belong to the same
    sequence, or set of sequences.

    `RepeatList` contains methods that act on several tandem repeats.
    For example, methods to

    *   detect overlapping tandem repeats
    *   identify the tandem repeat with highest statistical significance
    *   filter a set of tandem repeats according to different filtering
        procedures.


    Attributes:
        repeats (list of Repeat): The list of tandem repats.
    """

    def __str__(self):
        """
            Create string for RepeatList instance.
        """
        return str(self.repeats)

    def __init__(self, repeats):
        self.repeats = repeats

    def __add__(self, rl):
        if rl:
            return RepeatList(self.repeats + rl.repeats)
        else:
            return self

    def intersection(self, rl):
        if rl:
            return RepeatList([i for i in self.repeats if i in rl.repeats])
        else:
            return self

    def create(file, input_format):
        """ Read ``RepeatList`` from file.

        Read ``RepeatList`` from file (currently, only pickle is supported)

        Args:
            input_format (str):  Currently only "pickle"
            file (str): Path to output file

        .. todo:: Write checks for ``input_format`` and ``file``.

        """

        if input_format == 'pickle':
            with open(file, 'rb') as fh:
                return pickle.load(fh)
        else:
            raise Exception('input_format is unknown: {}.'.format(input_format))

    def write(self, output_format, file=None, return_string=None, *args):
        """ Serialize and write ``RepeatList`` instances.

        Serialize ``RepeatList`` instance using the stated ``output_format``.
        If a ``file`` is specified, save the String. If ``str`` is specified, give back
        the String (not possible for pickles).

        Args:
            output_format (str):  The output format: Either "pickle" or "tsv"
            file (str): Path to output file

        .. todo:: Write checks for ``output_format`` and ``file``.
        """

        if output_format == 'pickle':
            with open(file, 'wb') as fh:
                pickle.dump(self, fh)
            file = False
        elif output_format == 'tsv':
            output = repeat_list_io.serialize_repeat_list_tsv(self)
        else:
            raise Exception('output_format is unknown: {}'.format(output_format))

        if file:
            with open(file, 'w') as fh:
                fh.write(output)
        if return_string:
            return output

    def filter(self, func_name, *args, **kwargs):
        """ Filter ``repeats`` according to ``func_name``.

        Filter ``repeats`` according to ``func_name``

        Args:
            func_name (str): The name of a local filtering method.

        """

        # Check: is func_name a str, or is it a method? if it is a method, it
        # must be in dir(): ``funcname in dir`` == TRUE?

        func = getattr(sys.modules[__name__], func_name)
        return RepeatList(func(self, *args, **kwargs))

    def cluster(self, overlap_type, *args):
        """ Cluster ``repeats`` according to ``overlap_type``.

        Cluster ``repeats`` according to ``overlap_type``. We assume that overlap of
        repeats is transitive: If A overlaps with B, and B overlaps with C, (A,B,C) form
        one cluster.
        The attribute ``cluster`` is initiated to a dict:

            self.cluster = {"overlap_type1": [(0,2),(1)], "overlap_type2": [(0),(1), (2)]}

        In this toy example, the first and the third repeat in ``repeats`` cluster
        according to "overlap_type1", whereas no repeats cluster according to
        "overlap_type2".

        Args:
            overlap_type (str): The name of a local pairwise repeat overlapping method.

        """

        if not hasattr(self, 'd_cluster'):
            self.d_cluster = {}

        l_cluster = []

        l_repeat_indices = list(range(len(self.repeats)))
        while l_repeat_indices:
            i_cluster = [l_repeat_indices.pop()]
            l_repeat_in_cluster_check = i_cluster[:]

            while l_repeat_in_cluster_check:
                check = l_repeat_in_cluster_check.pop()
                remaining_indices = l_repeat_indices[:]
                for i in remaining_indices:
                    if two_repeats_overlap(
                            overlap_type,
                            self.repeats[check],
                            self.repeats[i]):
                        i_cluster.append(i)
                        l_repeat_in_cluster_check.append(i)
                        l_repeat_indices.remove(i)

            l_cluster.append(i_cluster)

        l_cluster = [set(i) for i in l_cluster]
        self.d_cluster[overlap_type] = l_cluster


# FILTERING
# Filter methods (func) are all defined in the same way: They return a subset
# of self.repeats, also as a list.
#  Return is a (list of `Repeat`).

def pvalue(rl, score, threshold):
    """ Returns all repeats in ``rl`` with a p-Value below a certain threshold.

    Returns all repeats in ``rl`` with a p-Value below a certain threshold.

    Args:
        rl (RepeatList): An instance of the RepeatList class.
        score (str): The type of score defines the pvalue that is used for filtering
        threshold (float): All repeats with a pvalue of type `score` above this threshold
            are filtered out.

    """

    threshold = float(threshold)

    res = []
    for i_repeat in rl.repeats:
        if i_repeat.pvalue(score) <= threshold:
            res.append(i_repeat)
    return res


def divergence(rl, score, threshold):
    """ Returns all repeats in ``rl`` with a divergence below a certain
        threshold.

    Returns all repeats in ``rl`` with a divergence below a certain threshold.

    Args:
        rl (RepeatList): An instance of the RepeatList class.
        score (str): The type of score defines the divergence that is used for
                    filtering
        threshold (float): All repeats with a divergence of type `score` above
                            this threshold are filtered out.
    """

    res = []
    for i_repeat in rl.repeats:
        if i_repeat.divergence(score) <= threshold:
            res.append(i_repeat)
    return res


def attribute(rl, attribute, type, threshold):
    """ Returns all repeats in ``rl`` with a attribute below (above) a certain
        threshold.

    Returns all repeats in ``rl`` with a attribute below (above) a certain
    threshold.

    Args:
        rl (RepeatList): An instance of the RepeatList class.
        attribute (str): The attribute of the Repeat instance
        type (str): Either "min" or "max"
        threshold (float): All repeats with an attribute value below (above)
                          this threshold are filtered out.
    """

    threshold = float(threshold)

    res = []
    for i_repeat in rl.repeats:
        value = getattr(i_repeat, attribute)
        if type == "min":
            if value >= threshold:
                res.append(i_repeat)
        elif type == "max":
            if value <= threshold:
                res.append(i_repeat)
        else:
            raise Exception(
                "type must either be 'min' or 'max'. Instead, it is {}.".format(type))
    return res


def none_overlapping_fixed_repeats(rl, rl_fixed, overlap_type):
    """ Returns all repeats in ``rl`` none-overlapping with ``rl_fixed``.

    Returns all repeats in ``rl`` none-overlapping with ``rl_fixed`` according
    to ``overlap``.

    Args:
        rl (RepeatList): An instance of the RepeatList class.
        rl (rl_fixed): A second instance of the RepeatList class.
        overlap (list): First list element: Name (str) of an overlap method in
                        repeat_list. All remaining elements are additional
                        arguments for this class.
    """

    res = []
    for i_repeat in rl.repeats:

        for i_repeat_fixed in rl_fixed.repeats:
            if two_repeats_overlap(
                    overlap_type,
                    repeat1=i_repeat,
                    repeat2=i_repeat_fixed):
                break
        else:
            res.append(i_repeat)

    return res


def none_overlapping(rl, overlap, l_criterion):
    """ Returns all none-overlapping repeats in ``rl``.

    Returns all none-overlapping repeats in ``rl``. Repeats are clustered
    according to ``overlap``. Of each cluster, only the best repeat is returned
    according to ``dCriterion``.

    Args:
        rl (RepeatList): An instance of the RepeatList class.
        overlap (tuple): First element: Name (str) of an overlap method in
            `repeat_list`. Second element: **kwargs. All remaining elements are
            additional arguments for this class.
        l_criterion (list): list of (criterion (str), criterion arguments)
            tuples. Until only one repeat is remainining in a cluster, the
            criteria are applied in order.
    """

    overlap_type = overlap[0]
    overlap_args = overlap[1]

    if not (hasattr(rl, 'd_cluster') and overlap_type in rl.d_cluster):
        rl.cluster(overlap_type, overlap_args)

    res = []
    for i_cluster in rl.d_cluster[overlap_type]:

        i_repeat = [rl.repeats[i] for i in i_cluster]

        for iC in l_criterion:
            criterion_type, criterion_value = iC
            if len(i_repeat) == 1:
                res.append(i_repeat[0])
                break

            if criterion_type == 'pvalue':
                min_value = min(i.pvalue(criterion_value) for i in i_repeat)
                i_repeat = [
                    i for i in i_repeat if i.pvalue(criterion_value) == min_value]

            elif criterion_type == 'divergence':
                min_value = min(i.divergence(criterion_value) for i in i_repeat)
                i_repeat = [
                    i for i in i_repeat if i.divergence(criterion_value) == min_value]

        else:
            LOG.debug(
                "repeat_list.none_overlapping(): > 1 Repeats have the same values...")
            res.append(i_repeat[0])

    return res


# OVERLAP DETECTION METHODS
def two_repeats_overlap(overlap_type, repeat1, repeat2):
    """ Helper method to test the overlap of ``repeat1`` and ``repeat2``.

    Helper method to test the overlap of ``repeat1`` and ``repeat2``. The
    overlap is calculated by the local method ``overlap_type``.

    Args:
        overlap_type (str): The name of a local pairwise repeat overlapping
                            method.
        repeat1 (Repeat): An instance of the Repeat class
        repeat2 (Repeat): A second instance of the Repeat class

    Returns:
        Forwards method output
    """

    is_overlapping = getattr(sys.modules[__name__], overlap_type)
    # Alternative implementation:
    # is_overlapping = globals()[overlap_type]
    return is_overlapping(repeat1, repeat2)


def shared_char(repeat1, repeat2):
    """ Do two TRs share at least one char?

    Return 1 if the two TRs share at least one char (amino acids or nucleotides); else 0.

    Args:
        repeat1 (Repeat): An instance of the Repeat class
        repeat2 (Repeat): A second instance of the Repeat class

    Returns:
        Bool: 1 if the repeats share >= 1 char, else 0.
    """

    if (repeat1.begin + repeat1.repeat_region_length - 1 < repeat2.begin) or \
       (repeat2.begin + repeat2.repeat_region_length - 1 < repeat1.begin):
        return False
    else:
        return True


def common_ancestry(repeat1, repeat2):
    """ Do two TRs share at least one pair of chars with common ancestry?

    Return 1 if the two TRs share at least one pair of chars (amino acids or
    nucleotides) with common ancestry; else 0.

    Args:
        repeat1 (Repeat): An instance of the Repeat class
        repeat2 (Repeat): A second instance of the Repeat class

    Returns:
        Bool: 1 if the repeats share >= 1 pair of chars with common ancestry, else 0.

    """

    if not hasattr(repeat1, 'msaIT'):
        repeat1.calc_index_msa()
    if not hasattr(repeat2, 'msaIT'):
        repeat2.calc_index_msa()

    original = repeat2.msaIT
    potential = repeat1.msaIT
    try:
        for p in potential:
            iP = 0
            for iP in range(len(p) - 1):
                i = 0
                j = 0
                while original[i][j] != p[iP]:
                    if original[i][j] > p[iP]:
                        i += 1
                        j = 0
                    else:
                        j += 1
                    if len(original) <= i or len(original[i]) <= j:
                        break
                else:
                    for iPRest in range(iP + 1, len(p)):
                        if p[iPRest] in original[i][j + 1:]:
                            #coverage = repeat1.repeat_region_length/repeat2.repeat_region_length
                            #greediness = repeat1.l_effective/repeat2.l_effective
                            return True
    except:
        LOG.warning('error in shared_char with original %s and potential %s',
                    str(repeat1.msaIT), str(repeat2.msaIT))
        LOG.warning('original:', str(repeat2.msa))
        LOG.warning('original:', str(repeat2.begin))
        LOG.warning('potential:', str(repeat1.msa))
        LOG.warning('potential:', str(repeat1.begin))
        return False
    return False



import os
import pytest

from tral.hmm.hmm import HMM
from tral.repeat import repeat

TEST_SEQUENCE = "GYEDEDEDRPFYALGLGKRPRTYSFGL"

TEST_REPEAT_MSA_DOUBLE = ["AA","AA"]
TEST_HMM_STATES_DOUBLE = ["N", "M1", "I1","M2","I2","C"]
TEST_HMM_P0_DOUBLE = {i:0 for i in TEST_HMM_STATES_DOUBLE}


TEST_REPEAT_MSA_SINGLE = ["A","A","A"]
TEST_HMM_STATES_SINGLE = ["N", "M1", "I1","C"]
TEST_HMM_P0_SINGLE = {i:0 for i in TEST_HMM_STATES_SINGLE}

# Test file names
TEST_FILE_WITH_ID = 'carcinustatin.hmm'

@pytest.fixture
def path():
    """Return the path to the test data files.
    """
    return os.path.join(os.path.abspath('.'), 'hmm', 'test')


def test_create_HMM_from_Repeat():

    test_repeat = repeat.Repeat(msa = TEST_REPEAT_MSA_DOUBLE)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)

    assert test_hmm.l_effective == 2
    assert set(test_hmm.states) == set(TEST_HMM_STATES_DOUBLE)
    assert test_hmm.p_0 == TEST_HMM_P0_DOUBLE
    #assert test_hmm.p_t == TEST_HMM_P0_DOUBLE


    test_repeat = repeat.Repeat(msa = TEST_REPEAT_MSA_SINGLE)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)

    assert test_hmm.l_effective == 1
    assert test_hmm.states == TEST_HMM_STATES_SINGLE
    assert test_hmm.p_0 == TEST_HMM_P0_SINGLE


def test_hmm_pickle():

    test_repeat = repeat.Repeat(msa = TEST_REPEAT_MSA_DOUBLE)
    test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat)

    test_pickle = os.path.join(path(), "test.pickle")
    test_hmm.write(test_pickle, 'pickle')
    test_hmm_new = HMM.create(input_format = 'pickle', file = test_pickle)

    assert test_hmm.hmmer == test_hmm_new.hmmer
    assert test_hmm.alphabet == test_hmm_new.alphabet

    if os.path.exists(test_pickle):
        os.remove(test_pickle)


import glob
import os
import shutil
import sys

try:
    from setuptools import setup, Command
except ImportError:
    from distutils.core import setup, Command

def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), "r") as f:
        return f.read()

# Set the home variable with user argument:
# Prettier solutions might be possible: http://stackoverflow.com/questions/677577/distutils-how-to-pass-a-user-defined-parameter-to-setup-py
try:
    i = sys.argv.index("--home")
    HOME = sys.argv[i + 1]
    del sys.argv[i+1]
    del sys.argv[i]
    if not os.path.exists(HOME):
        raise ValueError("The argument supplied in --home is not a valid path: {}".format(HOME))
except:
    HOME=os.path.expanduser("~")

SCRIPTS1 = [os.path.join("tral", "examples", i) for i in ["example_workflow_MBE2014.py"]]
SCRIPTS2 = [os.path.join("tral", "examples", "workflow", i) for i in ["tandem_repeat_annotation_scripts.py", "tandem_repeat_annotation_workflow.py"]]

setup(
    name="tral",
    version="0.3.5",
    author="Elke Schaper",
    author_email="elke.schaper@isb-sib.ch",
    packages=["tral", "tral.test", "tral.hmm", "tral.hmm.test", "tral.repeat", "tral.repeat.test", "tral.repeat_list", "tral.repeat_list.test", "tral.sequence", "tral.sequence.test"],
    #packages=find_packages(exclude=["tests*"]),
    scripts= SCRIPTS1 + SCRIPTS2,
    url="http://pypi.python.org/pypi/tral/",
    license="LICENSE.txt",
    description="Detect and evaluate tandem repeats in genomic sequence data.",
    long_description=read("README.rst"),
    #include_package_data=True, # If you want files mentioned in MANIFEST.in also to be installed, i.e. copied to usr/local/bin
    classifiers = [
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: OS Independent",
        ],
    install_requires=[
        "Biopython >= 1.64",
        "configobj >= 5.0.6",
        #"docutils >= 0.11", # Uncomment if you wish to create the documentation locally.
        "numpy >= 1.6.1",
        #"pypandoc >= 0.9.6" # Uncomment if you wish to convert the markdown readme to a rest readme for upload on Pypi.
        #"pytest >= 2.5.2", # Uncomment if you wish to run the tests locally.
        "scipy >=0.12.0",
        #"Sphinx >= 1.2.2", # Uncomment if you wish to create the documentation locally.
    ],
    # package_data: None-module files, which should still be distributed are mentioned here:
    package_data={"tral": ["tral_configuration/*.ini",
                           "tral_configuration/data/hhrepid/*",
                           "tral_configuration/data/substitution_rate_matrices/*",
                           "examples/*.py", "examples/data/*",
                           "examples/workflow/*.py", "examples/workflow/*.tsv",
                           "examples/workflow/*.ini", "examples/workflow/*.hmm",
                           "examples/workflow/*.fasta",
                           "examples/workflow/split_sequence_data/*.fasta"]},
    package_dir={"tral": "tral"},
)


TRAL = os.path.join(HOME, ".tral")
if os.path.exists(TRAL):
    print("The TRAL configuration directory {} already exists. The template configuration and datafiles are not copied to the already existing directory at this step.".format(TRAL))
else:
    shutil.copytree("tral/tral_configuration", TRAL)
    print("The TRAL configuration files and data files are now located in {}".format(TRAL))



# (C) 2015 Elke Schaper

"""
    :synopsis: Input/output for sequences

    .. moduleauthor:: Elke Schaper <elke.schaper@sib-sib.ch>
"""

from Bio import SeqIO
import logging

LOG = logging.getLogger(__name__)

#  ########## READ SEQUENCE ###################################################


def read_fasta(file, indices=None):
    """ Read all sequences from a fasta file.

    Read all sequences from a fasta file.
    At current, the Biopython SeqIO parser is used.

    Args:
        file (str): Path to input file
        start ([int, int]): Index of the first returned sequence, and the first
                            not returned sequence.

    """

    # Making a list out if the generator object might be overhead for huge
    # fastafiles
    count = 0
    for seq_record in SeqIO.parse(file, "fasta"):
        if indices:
            count += 1
            if count < indices[0]:
                continue
            elif count >= indices[1]:
                break
        yield str(seq_record.seq), seq_record.id


#   ########## WRITE SEQUENCE #################################################


def write(sequence, sequence_file, sequence_id="sequence_id_not_defined"):
    """ Write a sequence str to fasta format in specified <sequence_file>

    Write s sequence str to fasta format in specified <sequence_file>

    Args:
        sequence (str): Sequence
        sequence_file (str): Path to the output file
        sequence_id (str): ID of the sequence in the output file.

    """

    with open(sequence_file, 'a') as fastafile:
        fastafile.write(">" + str(sequence_id) + '\n')
        fastafile.write(str(sequence) + '\n')



import collections
import os
import pytest

from tral.repeat_list import repeat_list as rl
from tral.repeat import repeat
from tral.sequence import sequence

TEST_REPEATS = [["AA","AA"],["AAA","AAA"],["AAAA","AAAA"], ["AAA-","AAAA"]]
TEST_SCORE = "phylo_gap01"
TEST_SCORE_VALUE_LIST = [0.0, 0.5, 1.0,  1.0]
TEST_BEGIN_LIST = [6,10,10,10]
TEST_SEQUENCE = "MAAAAKAAAAAAL"

# The resulting string should contain the following data, however perhaps in a different order:
TEST_TSV = "msa_original\tbegin\tn_effective\tl_effective\tsequence_length\tpvalue\nAA,AA\t2\t2.0\t2\t4\tNone\nAAA,AAA\t7\t2.0\t3\t6\tNone"


@pytest.fixture
def path():
    """Return the path to the test data files.
    """
    return os.path.join(os.path.abspath('.'), 'repeat', 'test')


@pytest.mark.no_external_software_required
def test_create_repeat_list_from_repeats():

    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS]
    test_repeat_list = rl.RepeatList(repeats = test_repeats)

    assert len(test_repeat_list.repeats) == 4
    for i,j in zip (TEST_REPEATS, test_repeat_list.repeats):
        assert i == j.msa


@pytest.mark.no_external_software_required
def test_repeat_list_pickle():

    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS]
    test_repeat_list = rl.RepeatList(repeats = test_repeats)

    test_pickle = os.path.join(path(), "test.pickle")
    test_repeat_list.write('pickle', test_pickle)
    test_repeat_list_new = repeat.Repeat.create(test_pickle, 'pickle')

    assert len(test_repeat_list.repeats) == len(test_repeat_list_new.repeats)
    assert test_repeat_list.repeats[0].msa == test_repeat_list_new.repeats[0].msa

    if os.path.exists(test_pickle):
        os.remove(test_pickle)

def test_serialize_repeat_list_tsv():

    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS[:2]]
    test_seq = sequence.Sequence(TEST_SEQUENCE)
    for i in test_repeats:
        test_seq.repeat_in_sequence(i)
    test_repeat_list = rl.RepeatList(repeats = test_repeats)

    tsv = test_repeat_list.write("tsv", return_string = True)

    assert type(tsv) == str


@pytest.mark.no_external_software_required
def test_pairwise_overlap():

    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS]
    for i,j in zip(test_repeats, TEST_BEGIN_LIST):
        i.begin = j

    assert rl.two_repeats_overlap("common_ancestry", *test_repeats[:2]) == False
    assert rl.two_repeats_overlap("common_ancestry", *test_repeats[1:3]) == False
    assert rl.two_repeats_overlap("common_ancestry", *test_repeats[2:]) == False
    assert rl.two_repeats_overlap("shared_char", *test_repeats[:2]) == False
    assert rl.two_repeats_overlap("shared_char", *test_repeats[1:3]) == True


@pytest.mark.no_external_software_required
def test_cluster():

    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS]
    for i,j in zip(test_repeats, TEST_BEGIN_LIST):
        i.begin = j

    test_repeat_list = rl.RepeatList(repeats = test_repeats)
    test_repeat_list.cluster("common_ancestry")

    # Check whether both lists include exactly the same elements.
    for i in [{0}, {1,3}, {2}]:
        assert i in test_repeat_list.d_cluster["common_ancestry"]
    assert len(test_repeat_list.d_cluster["common_ancestry"]) == 3

    test_repeat_list.cluster("shared_char")

    # Check whether both lists include exactly the same elements.
    for i in [{0}, {1,2,3}]:
        assert i in test_repeat_list.d_cluster["shared_char"]
    assert len(test_repeat_list.d_cluster["shared_char"]) == 2


@pytest.mark.no_external_software_required
def test_filter_pvalue():

    #test_repeats = [repeat.Repeat(msa = i, scoreslist = ["phylo_gap01"], calc_score = True, calc_pvalue = True) for i in TEST_REPEATS]
    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS]
    for i,j in zip(test_repeats, TEST_SCORE_VALUE_LIST):
        i.d_pvalue = {}
        i.d_pvalue[TEST_SCORE] = j

    test_repeat_list = rl.RepeatList(repeats = test_repeats)

    test_repeat_list_filtered = test_repeat_list.filter("pvalue", TEST_SCORE, 0.1)
    assert len(test_repeat_list_filtered.repeats) == 1


@pytest.mark.no_external_software_required
def test_filter_cluster_based():

    test_repeats = [repeat.Repeat(msa = i) for i in TEST_REPEATS]
    for i,j in zip(test_repeats, TEST_SCORE_VALUE_LIST):
        i.d_pvalue = {}
        i.d_pvalue[TEST_SCORE] = j
    for i,j in zip(test_repeats, TEST_BEGIN_LIST):
        i.begin = j


    test_repeat_list = rl.RepeatList(repeats = test_repeats)
    test_repeat_list.filter("pvalue", TEST_SCORE, 0.1)
    test_repeat_list_filtered = test_repeat_list.filter("none_overlapping", ("common_ancestry", None), [("pvalue", TEST_SCORE), ("divergence", TEST_SCORE)])
    assert len(test_repeat_list_filtered.repeats) == 3
    for i in test_repeats[:3]:
        assert i in test_repeat_list_filtered.repeats


# (C) 2015 Elke Schaper

"""
    :synopsis: Input/output for the RepeatList class

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import logging

from tral import configuration

LOG = logging.getLogger(__name__)

CONFIG_GENERAL = configuration.Configuration.instance().config
CONFIG = CONFIG_GENERAL["repeat_list"]


def serialize_repeat_list_tsv(tandem_repeats, config=CONFIG, *args):
    ''' Serialize a ``repeat_list`` instance as tsv.

        config defines which tandem repeat characteristics are added to the
        .tsv. This could be for example:

        * begin: position of the tandem repeats within the sequence,
        * pvalue: statistical significance of the tandem repeats
        * divergence: divergence of the tandem repeat units
        * l_effective: length of the tandem repeat units
        * n_effective: number of tandem repeat units

    Attributes:
        tandem_repeats (repeat_list): A ``Repeat_list`` instance.
        config (dict): A dictionary. E.g.:
                        {"output_characteristics": ["n_effective", "pvalue"],
                         model: "phylo_gap01"}
        *args: Additional arguments

    Returns:
        str: The tsv as a string.

    '''

    output_characteristics = config["output_characteristics"]
    model = config["model"]

    data_all = ["\t".join(output_characteristics)]
    for i_repeat in tandem_repeats.repeats:
        data = []
        for i_characteristic in output_characteristics:
            if "divergence" == i_characteristic:
                if hasattr(i_repeat, "dScore") and model in i_repeat.dScore:
                    data.append(i_repeat.divergence(model))
                else:
                    data.append(None)
            elif "pvalue" == i_characteristic:
                if hasattr(i_repeat, "dPValue") and model in i_repeat.dPValue:
                    data.append(i_repeat.pvalue(model))
                else:
                    data.append(None)
            elif "score" == i_characteristic:
                if hasattr(i_repeat, "dScore") and model in i_repeat.dScore:
                    data.append(i_repeat.score(model))
                else:
                    data.append(None)
            elif "msa_original" == i_characteristic:
                try:
                    data.append(",".join(i_repeat.msa_original))
                except:
                    raise Exception("The attribute msa_original is not",
                                    "available for i_repeat.")
            elif getattr(i_repeat, i_characteristic):
                data.append(getattr(i_repeat, i_characteristic))
            else:
                raise Exception(
                    "The attribute {} is not available for tandem_repeats".format(i_characteristic))
        data_all.append("\t".join(str(i) for i in data))

    return "\n".join(data_all) + "\n"


def save_repeat_fasta(tandem_repeats, file):
    ''' save multiple <tandem_repeats> in Fasta format in specified <file>

        At current, only one TR per sequence can be defined, as the identifiers
        in the dict <tandem_repeats> must be unique.

        Attributes:
            tandem_repeats: Dict of tandem repeats and identifiers.
            e.g. {'ENSP00012': msa1, 'ENSP00013': msa2}

            >ID
            GHKI
            GHKI
            GH--
    '''

    with open(file, 'w', newline='\n') as fh:
        for identifier, msa in tandem_repeats.items():
            fh.write(">{0}\n".format(identifier))
            fh.write("\n".join(msa) + "\n\n")



# (C) 2011 Alexander Korsunsky
# (C) 2011-2015 Elke Schaper

"""
    :synopsis: The repeat score module.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import collections
import logging
import numpy as np
import os
from os.path import join
import re
import scipy.stats
import scipy.special
import scipy.linalg

from tral import configuration
from tral.paths import DATA_DIR

LOG = logging.getLogger(__name__)

CONFIG_GENERAL = configuration.Configuration.instance().config
CONFIG = CONFIG_GENERAL["repeat_score"]

# ############# REPEAT SCORE CALCULATION FUNCTIONS ############################
# ############# 1. SIMILARITY SCORES ##########################################


def load_equilibrium_freq(file_name):
    """ Load equilibrium frequencies from a substitution matrix file.

    Load equilibrium frequencies from a substitution matrix file. Note that
    with minor changes, the whole amino acid substitution matrix could be
    returned.

    Args:
        file_name (str): Path to substitution matrix file, E.g. LG.dat.

    Returns:
        (Dict of str, float): Dictionary of one letter amino acid codes and
                              float value
    """

    patstr_freqline = r"[+-]?\d+(?:\.\d+)?\s*"
    with open(file_name, "r") as f:
        # Create empty 20x20 matrix
        substitution_matrix = [[0.0] * 20] * 20
        row_count = 0
        for row in f:
            # print("row %d: " % row_count, row)
            row = row.strip()
            pat = r"\s*" + patstr_freqline * (row_count + 1)
            match = re.findall(pat, row)
            if not match:
                continue
            # store read lines into matrix
            substitution_matrix[row_count][0:row_count + 1] = \
                map(float, row.split())
            row_count = row_count + 1
            # Read 19 lines of substitution matrix
            if row_count == 19:
                break
        else:
            # EOF before reading the frequency line
            raise ValueError("Input file is malformed! Did not find enough",
                             "lines for the substitution matrix.")
        frequencies = []
        for row in f:
            matches = re.findall(patstr_freqline, row)
            frequencies.extend(map(float, matches))
            if len(frequencies) >= 20:
                break
        else:
            # EOF before finishing the frequency line
            raise ValueError("Input file is malformed! Did not find enough",
                             "numbers for equilibrium frequencies.")
        substitution_matrix[19][0:20] = frequencies[0:20]
    # return substitution_matrix
    return {aa: freq for aa,
            freq in zip(["A",
                         "R",
                         "N",
                         "D",
                         "C",
                         "Q",
                         "E",
                         "G",
                         "H",
                         "I",
                         "L",
                         "K",
                         "M",
                         "F",
                         "P",
                         "S",
                         "T",
                         "W",
                         "Y",
                         "V"],
                        substitution_matrix[19])}


def mean_similarity(repeat, measure_of_similarity,
                    ignore_gaps=CONFIG['indel'].as_bool('ignore_gaps')):
    """ Calculate the mean similarity of all columns in the multiple sequence
        alignment in ``repeat`` according to a ``measure_of_similarity``.

    Args:
        repeat (Repeat): An instance of the Repeat class.
        measure_of_similarity (str): Either "entropy", "parsimony", or "pSim".
        ignore_gaps (boolean): Are gaps, "-" treated as chars (False) or ignored
                              (True).

    Returns:
        float: A similarity measure for the repeat units in ``repeat``.

    """

    if not ignore_gaps:
        return sum(map(measure_of_similarity, repeat.msaTD)) / float(repeat.l)
    else:
        return sum(map(
            lambda column: measure_of_similarity(column.replace("-", "")),
            repeat.msaTD)) / repeat.l_effective  # python 2: integer division


def entropy(column):
    """ Calculate the entropy of a list.

    Calculate the entropy of list. Assume that the frequency of elements in
    the list reflects a probability density estimator.

    For an exact description see:

    Schaper, E., Kajava, A., Hauser, A., & Anisimova, M. Repeat or not repeat?
    --Statistical validation of tandem repeat prediction in genomic sequences.
    Nucleic Acids Research (2012).

    Args:
      column (list of str): List of elements that can be ordered in a set, e.g.
                            str or int.

    Returns:
        float: The entropy of ``column``.
    """

    estimatedP = [(column.count(a) / float(len(column))) for a in set(column)]
    return sum([-p * np.log2(p) for p in estimatedP if p > 0])


def parsimony(column):
    """ Calculate the parsimony score of a list.

    Calculate the parsimony score on ``column``:  Count the number of changes
    in a ``column`` of a multiple sequence alignment and divide by the
    maximally possible number of changes.

    For an exact description see:

    Schaper, E., Kajava, A., Hauser, A., & Anisimova, M. Repeat or not repeat?
    --Statistical validation of tandem repeat prediction in genomic sequences.
    Nucleic Acids Research (2012).

    Args:
      column (list of str): A ``list`` of two or more elements that can be
      ordered in a ``set``, e.g. ``str`` or ``int``

    Returns:
        float: The parsimony score of ``column``.

    ..  todo:: At which step shall we assert that the length of ``column`` > 1?
    """

    if len(column) > 1:
        # Here, 0 means no changes and 1 only changes # python 2: integer
        # division
        return (len(set(column)) - 1) / (len(column) - 1)
    else:
        LOG.warning("Repeats with n = 1 are not repeats")
        # raise AssertionError("Repeats with n = 1 are not repeats")
        return 1


def pSim(column):
    """ Calculate the pSim score of a list.

    Calculate the pSim score on ``column``:  Count the frequency of the most
    frequent element in a ``column`` of a multiple sequence alignment.

    For an exact description see:

    Schaper, E., Kajava, A., Hauser, A., & Anisimova, M. Repeat or not repeat?
    --Statistical validation of tandem repeat prediction in genomic sequences.
    Nucleic Acids Research (2012).

    Args:
      column (list of str): A list of elements.

    Returns:
        float: The pSim score of ``column``.
    """

    return collections.Counter(column).most_common(
        1)[0][1] / float(len(column))


#  #######  2. MAXIMUM LIKELIHOOD CALCULATIONS ################################


def optimisation(function, args,
                 start_min=CONFIG['optimisation'].as_float('start_min'),
                 start_max=CONFIG['optimisation'].as_float('start_max'),
                 n_iteration=CONFIG['optimisation'].as_int('n_iteration')):
    """Binary one-dimensional optimisation of ``function``.

    Perform a binary one dimensional optimisation of parameter ``t`` on
    ``function`` with the additional list of arguments/parameters ``args``.
    Start on a given range of start, and iterate ``n_iteration`` times. Return
    the maximum of the function, together with the maximum parameter ``t`` as a
    tuple.

    In addition to ``start_min`` and ``start_max``, additional hard boundaries
    on the optimisation parameter ``t`` are hard-coded: t E [0.0000000001, 100]

    Args:
      function (method): Method with input (float, ``args``) and output float.
      args (list of items): List of arguments for ``function``.

    Kwargs:
        start_min (float): Minimum value of function parameter ``t``.
        start_max (float): Maximum value of function parameter ``t``.
        n_iteration (int): Number of iterations until optimisation results are returned.

    Returns:
        tuple (Maximum of function: float,
               Corresponding function parameter: float)

      ..  todo:: Check why ``region_bounded`` is hardcoded at current.
      ..  todo:: Check why ``stepsize`` is hardcoded at current.
      ..  todo:: Generalise hard-coded optimisation parameter boundaries.
    """

    t = [[start_min], [(start_min + start_max) / 2], [start_max]]
    stepsize = 2
    for iT in t:
        iT.append(function(iT[0], *args))
    region_bounded = False
    count = 0
    while count < n_iteration:
        count += 1
        if not region_bounded:
            # First entry is best - shift trial out zone to left
            if t[0][1] > t[1][1]:
                t[2] = t[1][:]
                t[1] = t[0][:]
                t[0] = [max(0.0000000001, t[0][0] - stepsize)]
                t[0].append(function(t[0][0], *args))
            # Last entry is best - shift trial out zone to right
            elif t[2][1] > t[1][1]:
                t[0] = t[1][:]
                t[1] = t[2][:]
                t[2] = [min(100, t[2][0] + stepsize)]
                t[2].append(function(t[2][0], *args))
            else:
                region_bounded = True
        if region_bounded:
            # calc half_points
            half_points = [[(t[0][0] + t[1][0]) / 2], [(t[1][0] + t[2][0]) / 2]]
            for iT in half_points:
                iT.append(function(iT[0], *args))
            # First halfPoint is best - shift it to be the next midPoint
            if half_points[0][1] > t[1][1]:
                t[2] = t[1][:]
                t[1] = half_points[0][:]
            # Last halfPoint is best - shift it to be the next midPoint
            elif half_points[1][1] > t[1][1]:
                t[0] = t[1][:]
                t[1] = half_points[1][:]
            # Former midPoint is still the best - adjust the boundaries
            else:
                t[0] = half_points[0][:]
                t[2] = half_points[1][:]
    return (t[1][0], t[1][1])


def load_model(evolution_model='lg'):
    # Prepare the model of repeat evolution
    if evolution_model == 'lg':
        # 1. Substitutions are modeled by the LG matrix
        aa, eq_freq, substitution_matrix = LG()
    else:
        aa, eq_freq, substitution_matrix = K80()

    # equilibrium_freq = {i: j for i, j in zip(aa, eq_freq)}
    alphabet = {b: a for a, b in enumerate(aa)}
    length = len(aa)
    eq_freq = np.array(eq_freq)
    eq_freq = eq_freq / sum(eq_freq)
    S = np.array(substitution_matrix)
    PI = np.diag(eq_freq)
    Q = np.dot(S, PI)
    for i in range(length):
        Q[i, i] = - sum(Q[i][:i]) - sum(Q[i][i + 1:])
    diagonal = [Q[i, i] for i in range(length)]
    Q *= - sum((i * j for i, j in zip(diagonal, eq_freq)))
    return (Q, eq_freq, alphabet)


def TN93(alpha_1=CONFIG['TN93']['alpha_1'], alpha_2=CONFIG['TN93']['alpha_2'],
         beta=CONFIG['TN93']['beta']):
    '''TN93'''
    # 4*4 matrix with α1 = 0.3, α2 = 0.4, β = 0.7 according to TN93
    Q = [[-(alpha_1 + 2 * beta), alpha_1, beta, beta],
         [alpha_1, -(alpha_1 + 2 * beta), beta, beta],
         [beta, beta, -(alpha_2 + 2 * beta), alpha_2],
         [beta, beta, alpha_2, -(alpha_2 + 2 * beta)]]

    return ['T', 'C', 'A', 'G'], [0.25, 0.25, 0.25, 0.25], Q


def K80(kappa=CONFIG['K80']['kappa']):
    ''' K80 '''
    return TN93(alpha_1=kappa, alpha_2=kappa, beta=1)


def LG():
    '''LG'''
    return(['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M',
            'F', 'P', 'S', 'T', 'W', 'Y', 'V'],
           [0.079066, 0.055941, 0.041977, 0.053052, 0.012937, 0.040767,
            0.071586, 0.057337, 0.022355, 0.062157, 0.099081, 0.0646,
            0.022951, 0.042302, 0.04404, 0.061197, 0.053287, 0.012066,
            0.034155, 0.069147],
           [[-16.060388, 0.425093, 0.276818, 0.395144, 2.489084, 0.969894,
             1.038545, 2.06604, 0.358858, 0.14983, 0.395337, 0.536518,
             1.124035, 0.253701, 1.177651, 4.727182, 2.139501, 0.180717,
             0.218959, 2.54787], [0.425093, -21.193959, 0.751878, 0.123954,
            0.534551, 2.807908, 0.36397, 0.390192, 2.426601, 0.126991,
            0.301848, 6.326067, 0.484133, 0.052722, 0.332533, 0.858151,
            0.578987, 0.593607, 0.31444, 0.170887], [0.276818, 0.751878,
            -17.692284, 5.076149, 0.528768, 1.695752, 0.541712, 1.437645,
            4.509238, 0.191503, 0.068427, 2.145078, 0.371004, 0.089525,
            0.161787, 4.008358, 2.000679, 0.045376, 0.612025, 0.083688],
            [0.395144, 0.123954, 5.076149, -24.184967999999994, 0.062556,
             0.523386, 5.24387, 0.844926, 0.927114, 0.01069, 0.015076, 0.282959, 0.025548, 0.017416, 0.394456, 1.240275, 0.42586, 0.02989, 0.135107, 0.037967], [2.489084, 0.534551, 0.528768, 0.062556, -13.058626, 0.084808, 0.003499, 0.569265, 0.640543, 0.320627, 0.594007, 0.013266, 0.89368, 1.105251, 0.075382, 2.784478, 1.14348, 0.670128, 1.165532, 1.959291], [0.969894, 2.807908, 1.695752, 0.523386, 0.084808, -18.724285999999996, 4.128591, 0.267959, 4.813505, 0.072854, 0.582457, 3.234294, 1.672569, 0.035855, 0.624294, 1.223828, 1.080136, 0.236199, 0.257336, 0.210332], [1.038545, 0.36397, 0.541712, 5.24387, 0.003499, 4.128591, -23.8339, 0.348847, 0.423881, 0.044265, 0.069673, 1.807177, 0.173735, 0.018811, 0.419409, 0.611973, 0.604545, 0.077852, 0.120037, 0.245034], [2.06604, 0.390192, 1.437645, 0.844926, 0.569265, 0.267959, 0.348847, -16.613733, 0.311484, 0.008705, 0.044261, 0.296636, 0.139538, 0.089586, 0.196961, 1.73999, 0.129836, 0.268491, 0.054679, 0.076701], [0.358858, 2.426601, 4.509238, 0.927114, 0.640543, 4.813505, 0.423881, 0.311484, -6.179003, 0.108882, 0.366317, 0.697264, 0.442472, 0.682139, 0.508851, 0.990012, 0.584262, 0.597054, 5.306834, 0.119013], [0.14983, 0.126991, 0.191503, 0.01069, 0.320627, 0.072854, 0.044265, 0.008705, 0.108882, -22.217359, 4.145067, 0.159069, 4.273607, 1.112727, 0.078281, 0.064105, 1.033739, 0.11166, 0.232523, 10.649107], [0.395337, 0.301848, 0.068427, 0.015076, 0.594007, 0.582457, 0.069673, 0.044261, 0.366317, 4.145067, -31.146227, 0.1375, 6.312358, 2.592692, 0.24906, 0.182287, 0.302936, 0.619632, 0.299648, 1.702745], [0.536518, 6.326067, 2.145078, 0.282959, 0.013266, 3.234294, 1.807177, 0.296636, 0.697264, 0.159069, 0.1375, -19.90551, 0.656604, 0.023918, 0.390322, 0.748683, 1.136863, 0.049906, 0.131932, 0.185202], [1.124035, 0.484133, 0.371004, 0.025548, 0.89368, 1.672569, 0.173735, 0.139538, 0.442472, 4.273607, 6.312358, 0.656604, -24.724051, 1.798853, 0.099849, 0.34696, 2.020366, 0.696175, 0.481306, 1.898718], [0.253701, 0.052722, 0.089525, 0.017416, 1.105251, 0.035855, 0.018811, 0.089586, 0.682139, 1.112727, 2.592692, 0.023918, 1.798853, -23.599417, 0.094464, 0.361819, 0.165001, 2.457121, 7.803902, 0.654683], [1.177651, 0.332533, 0.161787, 0.394456, 0.075382, 0.624294, 0.419409, 0.196961, 0.508851, 0.078281, 0.24906, 0.390322, 0.099849, 0.094464, -16.74927, 1.338132, 0.571468, 0.095131, 0.089613, 0.296501], [4.727182, 0.858151, 4.008358, 1.240275, 2.784478, 1.223828, 0.611973, 1.73999, 0.990012, 0.064105, 0.182287, 0.748683, 0.34696, 0.361819, 1.338132, -12.953694999999998, 6.472279, 0.248862, 0.400547, 0.098369], [2.139501, 0.578987, 2.000679, 0.42586, 1.14348, 1.080136, 0.604545, 0.129836, 0.584262, 1.033739, 0.302936, 1.136863, 2.020366, 0.165001, 0.571468, 6.472279, -20.294897, 0.140825, 0.245841, 2.188158], [0.180717, 0.593607, 0.045376, 0.02989, 0.670128, 0.236199, 0.077852, 0.268491, 0.597054, 0.11166, 0.619632, 0.049906, 0.696175, 2.457121, 0.095131, 0.248862, 0.140825, -19.666723, 3.151815, 0.18951], [0.218959, 0.31444, 0.612025, 0.135107, 1.165532, 0.257336, 0.120037, 0.054679, 5.306834, 0.232523, 0.299648, 0.131932, 0.481306, 7.803902, 0.089613, 0.400547, 0.245841, 3.151815, -20.084989000000004, 0.249313], [2.54787, 0.170887, 0.083688, 0.037967, 1.959291, 0.210332, 0.245034, 0.076701, 0.119013, 10.649107, 1.702745, 0.185202, 1.898718, 0.654683, 0.296501, 0.098369, 2.188158, 0.18951, 0.249313, -18.608258]]
           )


def loglikelihood_substitution(t, Q, eq_freq, alphabet, tandem_repeat):
    """ Calculate the likelihood of a repeat assuming a star tree and a
        sequence model defined by ``Q``, ``t``, ``eq_freq`` and ``alphabet``.

    Args:
        t (float): The divergence of the ``tandem_repeat`` units.
        Q (dict ?): A substitution matrix.
        eq_freq (dict ?): Equilibrium frequencies of all chars.
        alphabet (dict of str,?):  An alphabet from a substitution matrix.
        tandem_repeat (Repeat): An instance of the ``Repeat`` class.

    Returns:
        float: The loglikelihood value for the tandem repeat.

    """

    P = scipy.linalg.expm(Q * t)

    # msaTDN is a list of tandem_repeat.n numpy arrays.
    # Each numpy array represents the elements of the alphabet. The entries are
    # the count of an alphabet element in the respective tandem_repeat column
    try:
        tandem_repeat.msaTDN
    except:
        tandem_repeat.msaTDN = []
        for column in tandem_repeat.msaTD_standard_aa:
            column = column.replace("-", "")
            my_diag = {i: 0 for i in range(len(alphabet.keys()))}
            for iN in column:
                my_diag[alphabet[iN]] = my_diag[alphabet[iN]] + 1
            tandem_repeat.msaTDN.append(
                np.array([j for j in my_diag.values()]))

    # First sum: Over all sites
    # Second sum: Over all possible ancestral characters
    # Third product: Over all repeat units
    likelihood = sum(np.log(np.sum(iJ * np.product(np.power(P[i], column))
                                   for i, iJ in enumerate(eq_freq)))
                     for column in tandem_repeat.msaTDN)
    return likelihood


def loglikelihood_gaps_starphylogeny_zipfian(
        t,
        tandem_repeat,
        indel_rate_per_site=CONFIG['indel'].as_float('indel_rate_per_site'),
        gaps=CONFIG['indel']['gaps'],
        indel_zipf=CONFIG['indel'].as_float('zipf')):
    """ Calculate the log-likelihood of the gap structure in a
        ``tandem repeat`` assuming a star tree.

    Calculate the log-likelihood of the gap structure in a ``tandem repeat``
    assuming a star tree (i.e. repeat unit correlations are not modeled). The
    gap length model is a zipfian distribution with parameter ``indel_zipf``.
    The gap generation model is a simple poisson model with rate parameter
    ``indel_rate_per_site``.

    Args:
        t (float): The divergence of the ``tandem_repeat`` units.
        tandem_repeat (Repeat): An instance of the ``Repeat`` class.
        indel_rate_per_site (float): The mutation rate of insertions and
         deletions, as opposed to ``t``.
        gaps (str): The mode of gap counting. Options are "row_wise",
        "ignore_coherent_deletions", "ignore_trailing_gaps" and
        "ignore_trailing_gaps_and_coherent_deletions".
        indel_zipf (float): The parameter of the Zipfian distribution.

    Returns:
        float: The loglikelihood value for the gap structure of the tandem
                repeat.

    ..  todo:: USE LOG EARLIER, NOT JUST IN RETURN STATEMENT
    """

    insertions = tandem_repeat.insertions
    deletions = tandem_repeat.deletions[gaps]
    gaps = tandem_repeat.gaps[gaps]

    # Calculate likelihood of the gap lengths with insertions and deletions
    # combined.
    # Here, we use the Zipfian distributions
    l_length = scipy.special.zeta(indel_zipf, 1) ** - len(gaps)
    for gap in gaps:
        l_length = l_length * (gap ** -indel_zipf)

    # Calculate the likelihood for the ratio of columns were an insertion has
    # taken place, as opposed to the number of columns, where no insertion has
    # taken place. Do the same for deletions.
    # Here, we use the Poisson model
    probability_gap_per_site = t * indel_rate_per_site / 2
    l_insertions = scipy.stats.distributions.binom.pmf(
        len(insertions),
        tandem_repeat.l_effective * tandem_repeat.n + 1,
        probability_gap_per_site)
    l_deletions = \
        scipy.stats.distributions.binom.pmf(len(deletions),
                                            tandem_repeat.l_effective * tandem_repeat.n,
                                            probability_gap_per_site)

    # Return the total likelihood of tandem_repeat's gap structure:
    return np.log(l_length * l_insertions * l_deletions)


def loglikelihood_substitutions_gaps(t, substitution_args, gap_args):
    return loglikelihood_substitution(t, *substitution_args) + \
        loglikelihood_gaps_starphylogeny_zipfian(t, *gap_args)


def loglikelihood_startopology_local(
        tandem_repeat,
        evolution_model=CONFIG['evolutionary_model'],
        gaps=CONFIG['indel']['gaps'],
        indel_rate_per_site=CONFIG['indel'].as_float('indel_rate_per_site'),
        parameters=False):

    if tandem_repeat.l_effective == 0:
        return 100, -100

    if parameters:
        Q, eq_freq, alphabet = parameters
    else:
        Q, eq_freq, alphabet = load_model(evolution_model)

    if gaps:
        divergence, loglikelihood_duplication_history = \
            optimisation(function=loglikelihood_substitutions_gaps,
                         args=[[Q, eq_freq, alphabet, tandem_repeat],
                               [tandem_repeat, indel_rate_per_site, gaps]])
    else:
        divergence, loglikelihood_duplication_history = \
            optimisation(function=loglikelihood_substitution,
                         args=[Q, eq_freq, alphabet, tandem_repeat])

    return divergence, loglikelihood_duplication_history


def loglikelihood_random(repeat,
                         evolution_model=CONFIG['evolutionary_model'],
                         parameters=False):

    if repeat.sequence_type == 'AA':
        if parameters:
            equilibrium_freq = parameters
        else:
            aarate_file_name = join(
                DATA_DIR,
                "substitution_rate_matrices",
                evolution_model +
                ".dat")  # load equilibrium frequencies
            equilibrium_freq = load_equilibrium_freq(aarate_file_name)

        loglikelihoodrandom = 0.
        for aa, freq in equilibrium_freq.items():
            # np.log uses the natural logarithm.
            loglikelihoodrandom += repeat.textD_standard_aa.count(
                aa) * np.log(float(freq))
    elif repeat.sequence_type == 'DNA':
        # CHECK THIS!
        # At the moment, we assume equal frequencies for each nucleotide -
        # Baseml Model J69 or K80
        # python2: float for division neccesary
        # np.log uses the natural logarithm.
        loglikelihoodrandom = repeat.totD * np.log(0.25)
    return loglikelihoodrandom


def phylo_star_topology_local(tandem_repeat, evolution_model=False, gaps=False,
                              indel_rate_per_site=0.001, parameters=False):
    if not evolution_model:
        if tandem_repeat.sequence_type == 'AA':
            evolution_model = 'lg'
        else:
            evolution_model = 'k80'

    divergence, logLH1 = \
        loglikelihood_startopology_local(tandem_repeat=tandem_repeat,
                                         evolution_model=evolution_model,
                                         gaps=gaps,
                                         indel_rate_per_site=indel_rate_per_site,
                                         parameters=parameters)
    logLH0 = loglikelihood_random(
        repeat=tandem_repeat,
        evolution_model=evolution_model)

    return divergence, 2 * (logLH1 - logLH0)


#    ####### TANDEM REPEAT SCORE CALIBRATION ##################################


def calc_score(
        repeats,
        result_path,
        result_file_name,
        scoreslist=CONFIG['score_calibration'].as_list('scoreslist'),
        save_calibration=CONFIG['score_calibration'].as_bool('save_calibration'),
        precision=CONFIG['score_calibration'].as_int('precision')):

    sequence_type = repeats[0].sequence_type
    if repeats[0].sequence_type == 'AA':
        evolution_model = 'lg'
    else:
        evolution_model = 'k80'
    Q, eq_freq, alphabet = load_model(evolution_model=evolution_model)
    parameters = [Q, eq_freq, alphabet]

    if sequence_type == 'AA':
        aarate_file_name = join(
            DATA_DIR,
            "substitution_rate_matrices",
            evolution_model +
            ".dat")  # load equilibrium frequencies
        equilibrium_freq = load_equilibrium_freq(aarate_file_name)
    else:
        equilibrium_freq = {i: 0.25 for i in ['A', 'C', 'G', 'T']}

    random = [
        loglikelihood_random(
            repeat=iRepeat,
            evolution_model=evolution_model,
            parameters=equilibrium_freq) for iRepeat in repeats]
    for i_score in scoreslist:
        if i_score == 'phylo':
            test_statistic = [
                loglikelihood_startopology_local(
                    iRepeat,
                    gaps=False,
                    parameters=parameters)[1] for iRepeat in repeats]
            test_statistic = [2 * (iT - iR) if iT != - 1 else - 1000 \
                              for iT, iR in zip(test_statistic, random)]
        elif i_score == 'phylo_gap01':
            test_statistic = \
            [loglikelihood_startopology_local(iRepeat, gaps=True,
                                              indel_rate_per_site=0.01,
                                              parameters=parameters)[1] \
            for iRepeat in repeats]
            test_statistic = [2 *(iT - iR) if iT != - 1 else - 1000 \
                              for iT, iR in zip(test_statistic, random)]
        elif i_score == 'phylo_gap001':
            test_statistic = \
                [loglikelihood_startopology_local(iRepeat, gaps=True,
                                                  indel_rate_per_site=0.001,
                                                  parameters=parameters)[1] \
                for iRepeat in repeats]
            test_statistic = [2 *(iT - iR) if iT != - 1 else - 1000 \
                              for iT, iR in zip(test_statistic, random)]
        elif i_score == 'phylo_paml':  # Is this defined anywhere?
            test_statistic = [
                loglikelihood_starTopology_paml(iRepeat) for iRepeat in repeats]
            test_statistic = [2 *(iT - iR) if iT != - 1 else - 1000 \
                              for iT, iR in zip(test_statistic, random)]
        elif i_score == 'parsimony':
            test_statistic = [np.round_(mean_similarity(iRepeat, parsimony),
                                        decimals=precision) if \
                              iRepeat.l_effective != 0 else -1 \
                              for iRepeat in repeats]
        elif i_score == 'pSim':
            test_statistic = [np.round_(mean_similarity(iRepeat, pSim),
                                        decimals=precision) if \
                              iRepeat.l_effective != 0 else -1 \
                              for iRepeat in repeats]
        elif i_score == 'entropy':
            test_statistic = [np.round_(mean_similarity(iRepeat, entropy),
                                        decimals=precision) if \
                              iRepeat.l_effective != 0 else -1 \
                              for iRepeat in repeats]

        if save_calibration:
            file_name = os.path.join(result_path, i_score, result_file_name)
            print(file_name)
            np.savez(file_name, np.sort(test_statistic))
        else:
            if not hasattr(repeats[0], 'score'):
                for iR in repeats:
                    iR.score = {}
            for iR, iT in zip(repeats, test_statistic):
                iR.dScore[i_score] = iT


def calibrate_score(
        repeats,
        result_file_path,
        scoreslist=CONFIG['score_calibration'].as_list('scoreslist')):

    for i_score in scoreslist:
        test_statistic = \
            np.sort([iRepeat.dScore[i_score] for iRepeat in repeats])
        np.savez(os.path.join('_'.join([result_file_path, i_score])),
                 test_statistic)

#  #######  CALCULATE AND SAVE THE PDF OF SCORES (Potentially deprecated) ###

def save_distribution(values, items, result_file_path, file_name, inverse):
    for iC in range(len(items)):
        val = np.array([i[iC] for i in values])
        result_val, inv_ndx = np.unique(val, return_inverse=True)
        result_p = np.bincount(inv_ndx)
        result_p = result_p / float(sum(result_p))
        if inverse and items[iC] in [
                'phylo',
                'pSim']:  # Are high values the best?
            result_p = result_p[::-1]
            result_val = result_val[::-1]
        result_p = result_p.cumsum()
        if not os.path.isdir(os.path.join(result_file_path, items[iC])):
            os.makedirs(os.path.join(result_file_path, items[iC]))
        np.savez(os.path.join(result_file_path, items[iC], file_name),
                 cdf=result_p, value=result_val)


def calculate_pdf_scores(
        repeats,
        result_file_path,
        file_name,
        scoreslist=CONFIG['score_calibration'].as_list('scoreslist')):
    """ CALCULATE THE PROBABILITY DISTRIBUTION OF SCORES """

    # Can you generalise the next command for arbitrary classifiers?
    scores = [
        [repeat.score(scoreslist[0]), repeat.score(scoreslist[1])] \
         for repeat in repeats]
    save_distribution(
        values=scores,
        items=scoreslist,
        result_file_path=result_file_path,
        file_name=file_name,
        inverse=True)



# (C) 2015 Elke Schaper

"""
    :synopsis: The repeat pvalue module.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""


import os
import bisect
from os.path import join
import csv
import logging
import numpy as np
import scipy as sp

from tral.paths import DATA_DIR
SCORE_DIR = join(DATA_DIR, 'pvalue')

LOG = logging.getLogger(__name__)

# ######################### REPEAT SCORE P-VALUE CALCULATION FUNCTIONS ########
# #################### phylo & entropy ########################################


def empirical_list(l, n, sequence_type='AA', score_type='phylo'):
    """Load and return a numpy list with 10,000 empirical values of the user defined
        distribution from an external file.


    If no distribution is available for a given ``l`` and ``n``, the closest values for
    ``l`` and ``n`` are chosen instead.

    For the standard model score "phylo", currently values are available until

        * ``l_max`` = 99
        * ``n_max`` = 49
        * ``total_repeat_length_max`` = 1000

    For ``l`` <=  ``l_max`` and ``n`` <=  ``n_max`` and ``n*l`` <=  ``total_repeat_length_max``
    all values are available.

    For other values, we use the closest distribution, assuming that values change little
    (heuristic assumption!).

    Args:
        l (int): The length of the repeat unit.
        n (int): The number of repeat units.
        sequence_type (str): The type of the sequence: either "AA" or "DNA".
        score_type: The model of repeat evolution used. E.g. "phylo".

    Returns:
        A numpy list of length 10.000


    .. todo:: Define "phylo" model.
    .. todo:: Perhaps add distributions beyond l_max, n_max.
    """

    l_max = 99
    n_max = 49
    if score_type == 'entropy':
        l_max = 20
        n_max = 5

    total_repeat_length_max = 1000
    l = min(l, l_max)
    n = min(n, n_max)
    if l * n > total_repeat_length_max:
        LOG.debug(
            "l: %d and n: %d are bigger than the total_repeat_length_max: %d" %
            (l, n, total_repeat_length_max))
        # Dirty little hack: as we do not have data for this pair of l and n,
        # apply nearest data file.
        while l * n > total_repeat_length_max:
            if l > n:
                l -= 1
            else:
                n -= 1

    file = join(SCORE_DIR, sequence_type, score_type,
                str(l) + '_' + str(n) + '.npz')
    if not os.path.isfile(file):
        raise ValueError("complete pdf file %s does not exist!" % file)

    empirical_list_all = np.load(file)
    # It is necessary to close the numpy filehandle.
    # Otherwise, there will be too many open operating system filehandles
    # if you run this function many times (more than ulimit -n allows, that is)
    empirical_list_part = empirical_list_all['arr_0']
    empirical_list_all.close()

    return empirical_list_part


def pvalue_from_empirical_list(tandemrepeat, score_type='phylo', score_value=None, empirical=[]):
    """Calculates the p-Value of a score_type for the given `tandemrepeat`.


    The p-Value is the number of scores for comparable tandem repeats, i.e. of
    same repeat unit length and repeat unit copy number that are as good or
    better.

    Args:
        tandemrepeat (Repeat): An instance of the Repeat class.
        score_type: The model of repeat evolution used. E.g. "phylo".
        score_value (float): The value of the score of the Repeat instance.
        empirical: The null distribution of the score for repeats of the same
                    type.

    Returns:
        pvalue: A float.

    """
    if score_value is None:
        score_value = tandemrepeat.score(score_type)

    if len(empirical) == 0:
        empirical = empirical_list(tandemrepeat.l_effective, tandemrepeat.n,
                                   tandemrepeat.sequence_type, score_type)

    # A smaller score is a better score for all scores in this list.
    if score_type in ['entropy']:
        if score_value == -1:
            return 1
        else:
            return bisect.bisect_right(empirical, score_value) / len(empirical)
    else:
        return 1 - bisect.bisect_left(empirical, score_value) / len(empirical)


# ####################### pSim & parsimony: read in PDF #######################

def column_pdf(n, score_type='psim', sequence_type='AA'):
    """Load and return the probability density function of the score on random
        ``sequence_type`` data of length ``n``.

    Load and return the probability density function of the score on random
    ``sequence_type`` data of length ``n``.
    This method handels scores with distributions independent of ``l``. This includes
    all parsimony scores, and excludes model based scores.

    The order of the values in the probability density function is WORST FIRST: The first
    values describe the probability for "bad" scores, the last values the probabilities
    for "top" scores.

    Currently, for all models values are available until

        * ``n_max`` = 150

    For ``n`` <=  ``n_max`` all values are available. If no distribution is available for
    a given ``n`` the closest values for ``n`` is used.

    For values of ``n`` above ``n_max``, the ``n_max`` pdf is returned.

    Args:
        n (int): The number of repeat units.
        score_type: The heuristic model used. E.g. "psim".
        sequence_type (str): The type of the sequence: either "AA" or "DNA".

    Returns:
        A numpy list of length 10.000

    .. todo:: Check return type.
    .. todo:: Check how this method concurs with the ``empirical_list`` method.
    """

    # Currently, the pdfs are only save up to n_max = 150
    n_max = 150
    with open(join(SCORE_DIR, sequence_type, score_type, str(min(n, n_max)) + '.txt'), 'r') as pdf_file:
        pdf = [i for i in csv.reader(pdf_file, dialect='excel-tab')]
        pdf = [np.double(i[0]) for i in pdf[1:]]
        return np.array(pdf)

# ############################# pSim & parsimony ##############################


def calculate_repeat_structure(tandemrepeat):
    """ Calculate the number of columns with a certain number of gaps for each column in
    a ``Repeat`` instance.

    You can use a different null distribution for each column of different length for both
    the parsimony and the pSim score. Therefore, calculate the structure of the TR before
    you calculate the null distribution of tandem repeat scores of tandem repeats with
    the same gap distribution.

    Args:
        n (int): The number of repeat units.
        score: The heuristic model used. E.g. "psim".
        sequence_type (str): The type of the sequence: either "AA" or "DNA".

    Returns:
        l_repeat_structure (list)
        n_repeat_structure (int)

    .. todo:: Include exact definition of returned values.
    """

    repeat_structure = [
        len(column.replace("-", ""))
        for column in tandemrepeat.msaTD if (2 * column.count('-') < len(column))
    ]
    n_repeat_structure = list(set(repeat_structure))
    l_repeat_structure = [
        repeat_structure.count(i) for i in n_repeat_structure
    ]
    return l_repeat_structure, n_repeat_structure


# #################### DERIVE p-Value distributions ###########################

def calc_pvalues(repeats, result_file_path, file_name,
                 scoretypes=['phylo', 'phylo_gap'], gappy_data=False):
    """ Create and save a null distribution for ``repeats`` scores for p-Value calculation.

    You can use a different null distribution for each column of different length for both
    the parsimony and the pSim score. Therefore, calculate the structure of the TR before
    you calculate the null distribution of tandem repeat scores of tandem repeats with
    the same gap distribution.

    Args:
        repeats (list of Repeat): A list of ``Repeat`` instances.
        result_file_path (str): Path to result folder.
        file_name (str): Name of result file.
        scoretypes (list of str):  List of scores. E.g. ['phylo','phylo_gap'].
        gappy_data (bool): True if any of ``repeats`` contain gaps, else False.

    """

    # If the repeats are not gappy, than you can use always the same distribution of scores
    # on random data to calculate the p-Value. Otherwise, there might be deletion columns
    # and the distribution should be loaded each time
    # 'pvalue_from_empirical_list' is called.
    empirical_list_local = []

    for i_score_type in scoretypes:
        if 'parsimony' == i_score_type:
            test_statistic = [
                pvalue_pars(i_repeat) if i_repeat.l_effective != 0 else 1 for i_repeat in repeats]
        elif 'pSim' == i_score_type:
            test_statistic = [
                pvalue_psim(i_repeat) if i_repeat.l_effective != 0 else 1 for i_repeat in repeats]
        else:
            if not gappy_data:
                empirical_list_local = empirical_list(
                    repeats[0].l_effective,
                    repeats[0].n,
                    repeats[0].sequence_type,
                    i_score_type)
            test_statistic = \
                [pvalue_from_empirical_list(i_repeat, i_score_type,
                                            empirical=empirical_list_local) \
                 if i_repeat.l_effective != 0 else 1 for i_repeat in repeats]
        score_path = os.path.join(result_file_path, i_score_type)
        if not os.path.isdir(score_path):
            os.makedirs(score_path)
        print(os.path.join(score_path, file_name))
        np.savez(os.path.join(score_path, file_name), np.sort(test_statistic))


def d_average_multinom(l, n, sequence_type, score_type):
    """ Helper function for the analytic calculation of parsimony or pSim scores.

    Called by ``d_average_multiple_max_multinom`` or ``d_average_multiple_pars_multinom``.

    Return the ``l``-times self-convoluted pdf of the score on random sequence_type data
    of length ``n``. The order of the pdf is kept from ``column_pdf()``.

    For details see:

    Schaper, E., Kajava, A., Hauser, A., & Anisimova, M. Repeat or not repeat?
    --Statistical validation of tandem repeat prediction in genomic sequences.
    Nucleic Acids Research (2012).

    Args:
        l (int): The length of the repeat unit.
        n (int): The number of repeat units.
        sequence_type (str): The type of the sequence: either "AA" or "DNA".
        score_type: The model of repeat evolution used. Either 'psim' or 'parsimony'.

    Returns:
        (description missing)

    .. warning:: if precision higher than max(uint32) use uint64 instead.
            CHECK: http://docs.scipy.org/doc/numpy/user/basics.types.html

    .. todo:: Describe return value.
    """

    complete_pdf = column_pdf(n=n, score_type=score_type, sequence_type=sequence_type)
    if l != 1:
        single_column_pdf = complete_pdf
        for i in range(2, l + 1):
            complete_pdf = sp.convolve(complete_pdf, single_column_pdf)
    return complete_pdf

# ####################### pSim ################################################


# python 2: precision must be float
def d_average_multiple_max_multinom(tandemrepeat, precision=10000.):
    """ Calculate null distribution for e.g. the pSim score for repeats of type ``tandemrepeat``
        as a probability density function.

    Analytically calculate the p-Value distribution for the pSim score for repeats of
    type ``tandemrepeat``. The derivation is described in:

    Schaper, E., Kajava, A., Hauser, A., & Anisimova, M. Repeat or not repeat?
    --Statistical validation of tandem repeat prediction in genomic sequences.
    Nucleic Acids Research (2012).

    Args:
        tandemrepeat (Repeat): A ``Repeat`` instance.
        precision (float): The precision of the returned probability density function in
        terms of the length of the resulting list.

    Returns:
        p (list of float): cumulated probabilities from 0 to 1.
        unnamed (list of float): scores corresponding to the probabilities in ``p``.

    .. warning:: if precision higher than max(uint32) use uint64 instead.
            CHECK: http://docs.scipy.org/doc/numpy/user/basics.types.html
    """

    l_repeat_structure, n_repeat_structure = calculate_repeat_structure(tandemrepeat)

    p = d_average_multinom(l_repeat_structure[0], n_repeat_structure[0],
                           tandemrepeat.sequence_type, score_type='psim')
    val = np.array((np.r_[l_repeat_structure[0]:(l_repeat_structure[0] * n_repeat_structure[0] + 1.)] * \
          (precision / n_repeat_structure[0])).round(), dtype='uint32')
    if len(l_repeat_structure) == 1:
        # return best values first. for pSim, best = 1, worst = 0
        return p[::-1], val[::-1] / (precision * sum(l_repeat_structure))
    else:
        for i in range(1, len(l_repeat_structure)):
            p = np.outer(p, d_average_multinom(l_repeat_structure[i],
                                               n_repeat_structure[i],
                                               tandemrepeat.sequence_type,
                                               score_type='psim')).ravel()
            val = (val[:, np.newaxis] +
                   np.array(
                (np.r_[l_repeat_structure[i]:(l_repeat_structure[i] * n_repeat_structure[i] + 1.)]
                 * (precision / n_repeat_structure[i])
                 ).round(),
                dtype='uint32')
            ).ravel()

            x = np.bincount(val, weights=p)
            val = np.unique(val)
            p = x[val]
        # return best values first. for pSim, best = 1, worst = 0
        return p[::-1], (val / (precision * sum(l_repeat_structure)))[::-1]


def pvalue_psim(tandemrepeat):
    """ Calculate the p-Value of the pSim score for a Repeat.

    Retrieve the probability density function for repeats of the same type as ``tandemrepeat``.
    Then, calculate the p-Value given this probability density function, and the
    pSim score of ``tandemrepeat``.

    Args:
        tandemrepeat (Repeat): A ``Repeat`` instance.

    Returns:
        p-Value (float)

    .. todo:: Check the method's behaviour if ``tandemrepeat`` s parsimony score has not been
        calculated before.
    .. todo:: Check exception: if pdf == False: return 1.
    .. todo:: Describe ``precision``.
    """

    precision = 10000.

    pdf = d_average_multiple_max_multinom(tandemrepeat, precision)
    cumsum_pdf = np.cumsum(pdf[0])

    #index = np.where(tandemrepeat.score('pSim') == pdf[1])[0]
    index = np.where(np.abs(tandemrepeat.score('pSim') - pdf[1]) <= 1. / precision)[0]
    if len(index) == 1:  # standard case: score is included in pdf[0]
        return(cumsum_pdf[index[0]])

    indices = np.where(tandemrepeat.score('pSim') > pdf[1])[0]
    if len(indices) >= 1:
        # if pars is not exactly included in list, give back mean of value above
        # and value below (should only occur due to numerical imprecision)
        a = min(indices)
        return((cumsum_pdf[a] + cumsum_pdf[a - 1]) / 2)
    # This should only occur if score('pSim') is really bad or really good
    else:
        return 1

# ################## parsimony ################################################


# python 2: precision must be float
def d_average_multiple_pars_multinom(tandemrepeat, precision=10000.):
    """ Calculate null distribution for the parsimony score for repeats of type ``tandemrepeat``
        as a probability density function.

    Analytically calculate the p-Value distribution for the parsimony score for repeats of
    type ``tandemrepeat``. The derivation is described in:

    Schaper, E., Kajava, A., Hauser, A., & Anisimova, M. Repeat or not repeat?
    --Statistical validation of tandem repeat prediction in genomic sequences.
    Nucleic Acids Research (2012).

    Args:
        tandemrepeat (Repeat): A ``Repeat`` instance.
        precision (float): The precision of the returned probability density function in
        terms of the length of the resulting list.

    Returns:
        p (list of float): cumulated probabilities from 0 to 1.
        unnamed (list of float): scores corresponding to the probabilities in ``p``.

    .. warning:: if precision higher than max(uint32) use uint64 instead.
            CHECK: http://docs.scipy.org/doc/numpy/user/basics.types.html
    """

    l_repeat_structure, n_repeat_structure = calculate_repeat_structure(tandemrepeat)

    p = d_average_multinom(
        l_repeat_structure[0], n_repeat_structure[0],
        tandemrepeat.sequence_type, score_type='parsimony'
    )
    val = np.array(
        (np.r_[(l_repeat_structure[0] * (n_repeat_structure[0] - 1)):-1:-1.]
         * (precision / (n_repeat_structure[0] - 1.))
         ).round(),
        dtype='uint32'
    )
    if len(l_repeat_structure) == 1:
        # return best values first. for parsimony, best = 0, worst = 1
        return p[::-1], (val / (precision * sum(l_repeat_structure)))[::-1]
    else:
        for i in range(1, len(l_repeat_structure)):
            try:
                p = np.outer(p,
                             d_average_multinom(
                                 l_repeat_structure[i],
                                 n_repeat_structure[i],
                                 tandemrepeat.sequence_type, score_type='parsimony')
                             ).ravel()

                val = (val[:, np.newaxis] + np.array(
                    (np.r_[(l_repeat_structure[i] * (n_repeat_structure[i] - 1)):-1:-1.]
                     * (precision / (n_repeat_structure[i] - 1.))
                     ).round(),
                    dtype='uint32')
                ).ravel()

                x = np.bincount(val, weights=p)
                val = np.unique(val)
                p = x[val]
            except:
                LOG.warning(
                    "Failed on: " + str(l_repeat_structure) + " " +
                    str(n_repeat_structure)
                )
                return False
        # return best values first. for parsimony, best = 0, worst = 1.
        # Here, the  np.bincount() and np.unique() funs did the reordering.
        return p, val / (precision * sum(l_repeat_structure))


def pvalue_pars(tandemrepeat):
    """ Calculate the p-Value of the parsimony score for a Repeat.

    Retrieve the probability density function for repeats of the same type as
    ``tandemrepeat``. Then, calculate the p-Value given this probability
    density function, and the parsimony score of ``tandemrepeat``.

    Args:
        tandemrepeat (Repeat): A ``Repeat`` instance.

    Returns:
        p-Value (float)

    .. todo:: Check the method's behaviour if ``tandemrepeat`` s parsimony
              score has not been calculated before.
    .. todo:: Check exception: if pdf == False: return 1.
    """

    precision = 10000.

    pdf = d_average_multiple_pars_multinom(tandemrepeat, precision)
    # Check the following three lines:
    if not pdf:
        return 1
    cumsum_pdf = np.cumsum(pdf[0])

    #index = np.where(pdf[1] == tandemrepeat.score('parsimony'))[0]
    index = np.where(
        np.abs(
            tandemrepeat.score('parsimony') -
            pdf[1]) <= 1. /
        precision)[0]
    if len(index) == 1:  # standard case: score is included in pdf[0]
        return cumsum_pdf[index[0]]

    indices = np.where(tandemrepeat.score('parsimony') < pdf[1])[0]
    if len(indices) >= 1:
        # if pars is not exactly included in list, give back mean of value
        # above and value below (should only occur due to numerical
        # imprecision)
        a = min(indices)
        return (cumsum_pdf[a] + cumsum_pdf[a - 1]) / 2
    else:  # This should only occur if score('parsimony') is really bad
        return 1


# #################### gap penalty ############################################

def gap_penalty(tandemrepeat, mu):
    """ Calculate the gap penalty for a ``Repeat`` given mutation rate ``mu``.

    Args:
        tandemrepeat (Repeat): A ``Repeat`` instance.
        mu (float): The mutation rate.

    Returns:
        Gap penalty (float)

    .. todo:: Define ``mu`` more precisely.
    .. todo:: Is this function called from anywhere? In case, consider refactoring.
    """

    return ((mu / (1 - mu)) ** tandemrepeat.n_gap_structure) * tandemrepeat.p_gap_structure
